//
//  Copyright (c) 2004-2006 by Autodesk, Inc.
//
//  By using this code, you are agreeing to the terms and conditions of
//  the License Agreement included in the documentation for this code.
//
//  AUTODESK MAKES NO WARRANTIES, EXPRESS OR IMPLIED, AS TO THE CORRECTNESS
//  OF THIS CODE OR ANY DERIVATIVE WORKS WHICH INCORPORATE IT. AUTODESK
//  PROVIDES THE CODE ON AN "AS-IS" BASIS AND EXPLICITLY DISCLAIMS ANY
//  LIABILITY, INCLUDING CONSEQUENTIAL AND INCIDENTAL DAMAGES FOR ERRORS,
//  OMISSIONS, AND OTHER PROBLEMS IN THE CODE.
//
//  Use, duplication, or disclosure by the U.S. Government is subject to
//  restrictions set forth in FAR 52.227-19 (Commercial Computer Software
//  Restricted Rights) and DFAR 252.227-7013(c)(1)(ii) (Rights in Technical
//  Data and Computer Software), as applicable.
//


#ifndef DWFTK_READ_ONLY


#include "dwfcore/MIME.h"
#include "dwfcore/Owner.h"
using namespace DWFCore;

#include "dwf/Version.h"
#include "dwf/package/XML.h"
#include "dwf/package/Manifest.h"
#include "dwf/package/Constants.h"
#include "dwf/package/EPlotSection.h"
#include "dwf/package/EModelSection.h"
#include "dwf/package/GlobalSection.h"
#include "dwf/package/ContentManager.h"
#include "dwf/package/utility/PropertyContainer.h"
#include "dwf/package/writer/extensions/6.0/PackageVersionExtension.h"
#include "dwf/package/writer/extensions/6.01/PackageVersionExtension.h"
#include "dwf/package/writer/extensions/6.11/PackageVersionExtension.h"
using namespace DWFToolkit;




    //
    // used for temporary buffer to transfer bytes
    // from a resource to the dwf archive
    //
#define _DWFTK_PACKAGE_WRITER_RESOURCE_STREAM_BUFFER_BYTES  16384

//DNT_Start

_DWFTK_API
DWFPackageWriter::DWFPackageWriter( const DWFFile&              rDWFPackageFile,
                                    const DWFString&            zDWFPackagePassword,
                                    DWFPackageVersionExtension* pVersionExtension )
throw()
                : _rDWFPackage( rDWFPackageFile )
                , _zDWFPassword( zDWFPackagePassword )
                , _pXMLSerializer( NULL )
                , _pVersionExtension( pVersionExtension )
                , _pPackageDescriptor( NULL )
                , _pPackageManifest( NULL )
                , _nNextPlot( 1.0 )
                , _bEPlotGlobalAdded( false )
                , _bEModelGlobalAdded( false )
                , _nEPlotSectionCount( 0 )
                , _nEModelSectionCount( 0 )
                , _oSections()
                , _oGlobalSections()
                , _bAggregating( false )
                , _eMergeType( eNoMerge )
                , _oSectionToContentManager()
                , _oContentToContentManager()
{;}

_DWFTK_API
DWFPackageWriter::~DWFPackageWriter()
throw()
{
    if (_pVersionExtension)
    {
        DWFCORE_FREE_OBJECT( _pVersionExtension );
    }

    if (_pXMLSerializer)
    {
        DWFCORE_FREE_OBJECT( _pXMLSerializer );
    }

    if (_pPackageDescriptor)
    {
        DWFCORE_FREE_OBJECT( _pPackageDescriptor );
    }

    if (_pPackageManifest)
    {
        DWFCORE_FREE_OBJECT( _pPackageManifest );
    }
}

_DWFTK_API
void
DWFPackageWriter::addSection( DWFSection*   pSection,
                              DWFInterface* pInterface )
throw( DWFException )
{
    if (_bAggregating)
    {
        //
        //  Grab the associated content manager before it gets swapped. 
        //  A later call to provideManifest will change the content manager on the section
        //  to this writer's associated content manager. 
        //
        DWFContentManager* pContentManager = pSection->getContentManager();
        if (pContentManager)
        {
            _oSectionToContentManager.insert( std::make_pair(pSection, pContentManager) );
        }
    }

    if (pSection == NULL)
    {
        _DWFCORE_THROW( DWFInvalidArgumentException, /*NOXLATE*/L"No section provided" );
    }

        //
        // always route through the version extension if it exists
        // and return immediately if instructed (ext. will return false)
        //
    if (_pVersionExtension && (_pVersionExtension->addSection(pSection, pInterface) == false))
    {
        return;
    }

        //
        // trap global sections
        //
    {
        DWFGlobalSection* pGlobal = dynamic_cast<DWFGlobalSection*>(pSection);
        if (pGlobal)
        {
            return addGlobalSection( pGlobal );
        }
    }

    //
    //
    //
    _ensureManifest();

        //
        // ask the section for it's interface (it may not exist)
        //
    if (pInterface == NULL)
    {
        pInterface = pSection->buildInterface();
    }

        //
        // 
        //
    if (pInterface)
    {
        //
        // add to the manifest
        //
        _pPackageManifest->provideInterface( pInterface );
    }

    //
    // Grab the behavior from the section. In 7.1, this was on CustomSection.
    // In 7.2, we've promoted it to Section in general.
    //
    DWFSection::tBehavior tSectionBehavior = pSection->behavior();

        //
        // modify the section before caching
        //
    if (tSectionBehavior.bRenameOnPublish)
    {
        pSection->rename( _oUUID.next(true) );
    }

        //
        // manage section descriptor resource
        //
    if (tSectionBehavior.bPublishDescriptor)
    {
        pSection->reorder( _nNextPlot++ );

        //
        // the HREF of this descriptor
        //
        DWFString zHREF( pSection->name() );
        zHREF.append( /*NOXLATE*/L"\\descriptor.xml" );

        //
        // create the descriptor resource
        //
        DWFResource* pDescriptor = DWFCORE_ALLOC_OBJECT( DWFResource(/*NOXLATE*/L"",
                                                            DWFXML::kzRole_Descriptor,
                                                            DWFMIME::kzMIMEType_XML,
                                                            zHREF) );
        if (pDescriptor == NULL)
        {
            _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate desriptor resource" );
        }

        //
        // remove any previous descriptor
        // and have it destroyed if the section owns it
        //
        pSection->removeResourcesByRole( DWFXML::kzRole_Descriptor, true );

        //
        // add a descriptor resource to the section
        //
        pSection->addResource( pDescriptor, true );
    }

    //
    // add the section to the manifest
    //
    _pPackageManifest->provideSection( pSection );

    //
    // cache the section
    //
    _oSections.push_back( pSection );

    //
    // increment the section specific count based on the section type
    //
    const DWFString& zSectionType = pSection->type();

    if (zSectionType == _DWF_FORMAT_EPLOT_TYPE_STRING)
    {
        _nEPlotSectionCount++;
    }
    else if (zSectionType == _DWF_FORMAT_EMODEL_TYPE_STRING)
    {
        _nEModelSectionCount++;
    }
}

_DWFTK_API
void
DWFPackageWriter::addGlobalSection( DWFGlobalSection* pSection )
throw( DWFException )
{
    if (pSection == NULL)
    {
        _DWFCORE_THROW( DWFInvalidArgumentException, /*NOXLATE*/L"No section provided" );
    }

        //
        // always route through the version extension if it exists
        // and return immediately if instructed (ext. will return false)
        //
    if (_pVersionExtension && (_pVersionExtension->addGlobalSection(pSection) == false))
    {
        return;
    }

    //
    //
    //
    _ensureManifest();

        //
        // ensure only one global section per type
        //
    if (pSection->type() == _DWF_FORMAT_EPLOT_GLOBAL_TYPE_STRING)
    {
        if (_bEPlotGlobalAdded)
        {
            DWFString zEx( /*NOXLATE*/L"Global section already exists [" );
                      zEx.append( _DWF_FORMAT_EPLOT_GLOBAL_TYPE_STRING );
                      zEx.append( /*NOXLATE*/L"]" );

            _DWFCORE_THROW( DWFUnexpectedException, zEx );
        }

        _bEPlotGlobalAdded = true;
    }
    else if (pSection->type() == _DWF_FORMAT_EMODEL_GLOBAL_TYPE_STRING)
    {
        if (_bEModelGlobalAdded)
        {
            DWFString zEx( /*NOXLATE*/L"Global section already exists [" );
                      zEx.append( _DWF_FORMAT_EMODEL_GLOBAL_TYPE_STRING );
                      zEx.append( /*NOXLATE*/L"]" );

            _DWFCORE_THROW( DWFUnexpectedException, zEx );
        }

        _bEModelGlobalAdded = true;
    }

    //
    // the HREF of this descriptor
    //
    DWFString zHREF( pSection->name() );
    zHREF.append( /*NOXLATE*/L"\\descriptor.xml" );

    //
    // create the descriptor resource
    //
    DWFResource* pDescriptor = DWFCORE_ALLOC_OBJECT( DWFResource(/*NOXLATE*/L"",
                                                        DWFXML::kzRole_Descriptor,
                                                        DWFMIME::kzMIMEType_XML,
                                                        zHREF) );
    if (pDescriptor == NULL)
    {
        _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate desriptor resource" );
    }

    //
    // remove any previous descriptor
    // and have it destroyed if the section owns it
    //
    pSection->removeResourcesByRole( DWFXML::kzRole_Descriptor, true );

    //
    // add a descriptor resource to the section
    //
    pSection->addResource( pDescriptor, true );

    //
    // add the section to the manifest
    //
    _pPackageManifest->provideSection( pSection );

    //
    // cache the section
    //
    _oGlobalSections.push_back( pSection );
}

_DWFTK_API
DWFInterface*
DWFPackageWriter::addInterface( DWFInterface* pInterface )
throw( DWFException )
{
    if (pInterface == NULL)
    {
        _DWFCORE_THROW( DWFInvalidArgumentException, /*NOXLATE*/L"Interface must be specified" );
    }

    //
    //
    //
    _ensureManifest();

    //
    // add to the manifest
    //
    return _pPackageManifest->provideInterface( pInterface );
}

_DWFTK_API
void 
DWFPackageWriter::attachContentManager( DWFContentManager* pContentManager, 
                                        bool bTakesOwnership )
throw( DWFException )
{
    _ensureManifest();

    _pPackageManifest->attachContentManager( pContentManager, bTakesOwnership );
}

_DWFTK_API
DWFContentManager*
DWFPackageWriter::getContentManager()
throw( DWFException )
{
    _ensureManifest();

    DWFContentManager* pContentManager = _pPackageManifest->getContentManager();
    return pContentManager;
}

_DWFTK_API
DWFPropertyContainer&
DWFPackageWriter::getManifestProperties()
throw( DWFException )
{
    //
    //
    //
    _ensureManifest();

    return dynamic_cast<DWFPropertyContainer&>(*_pPackageManifest);
}
    
_DWFTK_API
void
DWFPackageWriter::write( const DWFString&                   zSourceProductVendor,
                         const DWFString&                   zSourceProductName,
                         const DWFString&                   zSourceProductVersion,
                         const DWFString&                   zDWFProductVendor,
                         const DWFString&                   zDWFProductVersion,
                         DWFZipFileDescriptor::teFileMode   eCompressionMode )
throw( DWFException )
{
        //
        // Try to determine a typeinfo type.
        //
        // if no ePlot sections have been added write out eModel typeinfo
        // else write out the ePlot typeinfo
        //
    DWFString zTypeInfo;
    if (_nEPlotSectionCount == 0)
    {
        // No ePlot
        if (_nEModelSectionCount != 0)
        {
            // No ePlot and some eModel - tag as eModel.
            zTypeInfo = DWFInterface::kzEModel_ID;
        }
        else
        {
            // No ePlot and no emodel - do not create typeinfo.
            zTypeInfo = /*NOXLATE*/L"";
        }
    }
    else
    {
        // Some ePlot.
        if (_nEModelSectionCount != 0)
        {
            // Some ePlot, and some eModel - tag as ePlot
            zTypeInfo = DWFInterface::kzEPlot_ID;
        }
        else
        {
            // Some ePlot and no emodel.
            //
            // Due to viewer weirdness (pali reading a dwf > 6.0), we'll need to tag this as having typeinfo.
            zTypeInfo = DWFInterface::kzEPlot_ID;
        }
    }

        //
        // if we are aggregating, do this before using or serializing the content
        // information.
        //
    if (_bAggregating)
    {
        _aggregateContent();
    }

        //
        // if no version extension has been provided, hook in the appropriate
        // one based on the section type counts and by looking for content,
        // content presentations, or Raster Overlays.
        //
    if (_pVersionExtension == NULL)
    {

        //
        // First, check for Content. This came onboard in 6.20.
        //
        // If the ContentManager has anything in it, that means we have content.
        //
        bool bHaveContent = false;
        DWFContentManager* pContentManager = _pPackageManifest->getContentManager();
        if (pContentManager)
        {
            bHaveContent = (pContentManager->getContentCount() > 0);
        }

            //
            // If we don't have content, check for Content Presentation,
            // or Raster Overlay.
            // These also came onboard in 6.20.
            //
            // This check is more complex, as we have to go through all the
            // sections, looking for resources of type kzRole_ContentPresentation,
            // kzRole_MarkupContentPresentation, or kzRole_RasterOverlay.
            //
            // TODO: When we have global presentation data, we'll need to
            // detect it here as well.
            //
        bool bHavePresentation = false;
        bool bHaveRasterOverlay = false;
        if(bHaveContent == false)
        {
            DWFSection *pSection;
            DWFSection::tList::const_iterator iSection;
            for (iSection = _oSections.begin();
                iSection!= _oSections.end();
                iSection++)
            {
                pSection = *iSection;

                //
                // iterator for finders
                //
                DWFResourceContainer::ResourceKVIterator* piResourcesByRole = pSection->getResourcesByRole();
                if (piResourcesByRole && piResourcesByRole->valid())
                {
                    for(; piResourcesByRole->valid(); piResourcesByRole->next())
                    {
                        const DWFString zRole = piResourcesByRole->key();

                        if (zRole == DWFXML::kzRole_ContentPresentation ||
                            zRole == DWFXML::kzRole_MarkupContentPresentation)
                        {
                            bHavePresentation = true;
                            break;
                        }

                        if(zRole == DWFXML::kzRole_RasterOverlay)
                        {
                            bHaveRasterOverlay = true;
                            break;
                        }
                    }

                }

                DWFCORE_FREE_OBJECT( piResourcesByRole );
                    //
                    // If we've found one, that's all we need to know.
                    //
                if( (bHavePresentation == true) || (bHaveRasterOverlay == true) )
                {
                    break;
                }
            }
        }

            //
            // If we've detected content, presentation, or raster overlay, then we're
            // going to go with the present "default" behavior. Else, we will try
            // to write out an earlier version.
            //
        if( (bHaveContent == false) && (bHavePresentation == false) && (bHaveRasterOverlay == false) )
        {
                //
                // if no eModel sections have been added, use the 6.0 extension
                //
            if (_nEModelSectionCount == 0)
            {
                _pVersionExtension = DWFCORE_ALLOC_OBJECT( DWFPackageVersion60Extension() );
            }

                //
                // if only one eModel section has been added, use the TypeInfo extension
                //
            else if (_nEPlotSectionCount == 0 && _nEModelSectionCount == 1)
            {    
                _pVersionExtension = DWFCORE_ALLOC_OBJECT( DWFPackageVersionTypeInfoExtension(DWFPackageVersionTypeInfoExtension::kzDocumentType_EModel) );
            }

                //
                // else, we should output 6.11 - which was the first true "eComposite" version,
                // and may or may not have typeinfo (based on the argument to the constructor).
                //
            else
            {    
                _pVersionExtension = DWFCORE_ALLOC_OBJECT( DWFPackageVersion611Extension(zTypeInfo) );
            }
        }
    }

        //
        // always route through the version extension if it exists
        // and return immediately if instructed (ext. will return false)
        //
    if (_pVersionExtension && 
       (_pVersionExtension->write(zSourceProductVendor,
                                  zSourceProductName,
                                  zSourceProductVersion,
                                  zDWFProductVendor,
                                  zDWFProductVersion,
                                  eCompressionMode) == false))
    {
        return;
    }

    if (_pPackageManifest == NULL)
    {
        _DWFCORE_THROW( DWFNullPointerException, /*NOXLATE*/L"Cannot write package - no manifest exists" );
    }

    if (_pXMLSerializer == NULL)
    {
        _pXMLSerializer = DWFCORE_ALLOC_OBJECT( DWFXMLSerializer(_oUUID) );
        if (_pXMLSerializer == NULL)
        {
            _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate serializer" );
        }
    }

        //
        // open the package
        //
    if (_pPackageDescriptor == NULL)
    {
        if (_pVersionExtension)
        {
            _pPackageDescriptor = DWFCORE_ALLOC_OBJECT( DWFPackageFileDescriptor(_rDWFPackage, eCompressionMode, _pVersionExtension->major(), _pVersionExtension->minor()) );
        }
        else
        {
            _pPackageDescriptor = DWFCORE_ALLOC_OBJECT( DWFPackageFileDescriptor(_rDWFPackage, eCompressionMode) );
        }

        if (_pPackageDescriptor == NULL)
        {
            _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate file descriptor" );
        }

        _pPackageDescriptor->open();
    }

        //
        // if no version extension has been hooked in then write out a typeinfo file
        // 
    if (_pVersionExtension == NULL)
    {
        if (zTypeInfo.chars())
        {
            zTypeInfo.append( /*NOXLATE*/L".TYPEINFO" );
        
            //
            // create the file
            //
            DWFOutputStream* pFilestream = _pPackageDescriptor->zip( zTypeInfo, _zDWFPassword );

            //
            // this file is empty, so just delete the stream (and close the file)
            //
            DWFCORE_FREE_OBJECT( pFilestream );
        }
    }

        //
        // give extension chance to write a file before the manifest
        //
    if (_pVersionExtension)
    {
       _pVersionExtension->prewriteManifest( *this, *_pPackageDescriptor, *_pXMLSerializer, _zDWFPassword );
    }
    
    //
    // create the MANIFEST
    //
    DWFOutputStream* pFilestream = _pPackageDescriptor->zip( /*NOXLATE*/L"manifest.xml", _zDWFPassword );

    //
    // just before writing out the manifest,
    // add any document-wide properties to it
    //
    DWFProperty* pStandardProperty = NULL;

    {
        if (zSourceProductVendor.bytes() > 0)
        {
            pStandardProperty = DWFCORE_ALLOC_OBJECT( DWFProperty(/*NOXLATE*/L"SourceProductVendor", zSourceProductVendor) );
            if (pStandardProperty == NULL)
            {
                _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate property" );
            }

            _pPackageManifest->addProperty( pStandardProperty, true );
        }

        if (zSourceProductName.bytes() > 0)
        {
            pStandardProperty = DWFCORE_ALLOC_OBJECT( DWFProperty(/*NOXLATE*/L"SourceProductName", zSourceProductName) );
            if (pStandardProperty == NULL)
            {
                _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate property" );
            }

            _pPackageManifest->addProperty( pStandardProperty, true );
        }

        if (zSourceProductVersion.bytes() > 0)
        {
            pStandardProperty = DWFCORE_ALLOC_OBJECT( DWFProperty(/*NOXLATE*/L"SourceProductVersion", zSourceProductVersion) );
            if (pStandardProperty == NULL)
            {
                _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate property" );
            }

            _pPackageManifest->addProperty( pStandardProperty, true );
        }

        if (zDWFProductVendor.bytes() > 0)
        {
            pStandardProperty = DWFCORE_ALLOC_OBJECT( DWFProperty(/*NOXLATE*/L"DWFProductVendor", zDWFProductVendor) );
            if (pStandardProperty == NULL)
            {
                _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate property" );
            }

            _pPackageManifest->addProperty( pStandardProperty, true );
        }

        if (zDWFProductVersion.bytes() > 0)
        {
            pStandardProperty = DWFCORE_ALLOC_OBJECT( DWFProperty(/*NOXLATE*/L"DWFProductVersion", zDWFProductVersion) );
            if (pStandardProperty == NULL)
            {
                _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate property" );
            }
            _pPackageManifest->addProperty( pStandardProperty, true );
        }

        pStandardProperty = DWFCORE_ALLOC_OBJECT( DWFProperty(/*NOXLATE*/L"DWFToolkitVersion", _DWFTK_VERSION_STRING) );
        if (pStandardProperty == NULL)
        {
            _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to allocate property" );
        }
        _pPackageManifest->addProperty( pStandardProperty, true );
    }

    //
    // bind the stream to the writer, this will directly
    // populate the archive as the manifest serializes itself
    //
    _pXMLSerializer->attach( *pFilestream );

    //
    // tell the manifest to serialize himself as an XML document
    //
    _pPackageManifest->serializeXML( *_pXMLSerializer, eManifest );

    //
    // unbind the stream and finalize the file in the archive
    //
    _pXMLSerializer->detach();

    //
    // release the file stream
    //
    DWFCORE_FREE_OBJECT( pFilestream );

        //
        // give extension chance to write a file after the manifest
        //
    if (_pVersionExtension)
    {
       _pVersionExtension->postwriteManifest( *this, *_pPackageDescriptor, *_pXMLSerializer, _zDWFPassword );
    }

    //
    // Serialize all the contents in the content manager to the package.
    //
    DWFContentManager* pContentManager = _pPackageManifest->getContentManager();
    if (pContentManager)
    {
        DWFString zHREF;
        DWFContent::tMap::Iterator* piContents = pContentManager->getContents();

        if (piContents)
        {
            DWFContent* pContent = NULL;
            for (; piContents->valid(); piContents->next())
            {
                pContent = piContents->value();

                if (pContent->isLoaded())
                {
                   zHREF.assign( pContent->href() );

                    //
                    //  Start the content file.
                    //
                    pFilestream = _pPackageDescriptor->zip( zHREF, _zDWFPassword );
                    //
                    //  Bind the stream to the writer/serializer.
                    //
                    _pXMLSerializer->attach( *pFilestream );

                    //
                    //  Tell the content to serialize itself.
                    //
                    pContent->serializeXML( *_pXMLSerializer, eGlobalContent );

                    //
                    //  Unbind the stream. This will also finalize the file in the archive.
                    //
                    _pXMLSerializer->detach();
                    //
                    //  Then release unbound stream.
                    //
                    DWFCORE_FREE_OBJECT( pFilestream );
                }
                else
                {
                    //
                    //  Since the content isn't loaded, instead of loading it, lets try to get the input stream
                    //  directly write out the bytes to new package
                    //
                    _serializeContent( pContent );
                }
            }

            DWFCORE_FREE_OBJECT( piContents );
        }
    }

    //
    // Add any global sections first
    //
    DWFString zSection;
    DWFSection* pSection = NULL;
    DWFGlobalSection::tList::const_iterator iGlobalSection;

    for (iGlobalSection = _oGlobalSections.begin();
         iGlobalSection!= _oGlobalSections.end();
         iGlobalSection++)
    {
        pSection = *iGlobalSection;

        //
        // name the new descriptor
        //
        zSection.assign( pSection->name() );
        zSection.append( /*NOXLATE*/L"\\descriptor.xml" );

        //
        // start with the descriptor
        //
        pFilestream = _pPackageDescriptor->zip( zSection, _zDWFPassword );

        //
        // bind the stream to the writer, this will directly
        // populate the archive as the page descriptor serializes itself
        //
        _pXMLSerializer->attach( *pFilestream );

        //
        // tell the page to serialize itself into the descriptor XML document
        //
        pSection->serializeXML( *_pXMLSerializer, eDescriptor );

        //
        // unbind the stream and finalize the file in the archive
        //
        _pXMLSerializer->detach();

        //
        // release the file stream
        //
        DWFCORE_FREE_OBJECT( pFilestream );

        //
        // resources
        //
        DWFResource* pResource = NULL;
        DWFKVIterator<const wchar_t*, DWFResource*>* piAllResources = pSection->getResourcesByRole();

        if (piAllResources)
        {
            for (; piAllResources->valid(); piAllResources->next())
            {
                pResource = piAllResources->value();

                if (pResource->role() != DWFXML::kzRole_Descriptor)
                {
                    _serializeResource( pSection->name(), *pResource );
                }
            }

            DWFCORE_FREE_OBJECT( piAllResources );
        }
    }

    DWFSection::tBehavior tSectionBehavior;

        //
        // Add the non-global sections
        //
    DWFSection::tList::const_iterator iSection;
    for (iSection = _oSections.begin();
         iSection!= _oSections.end();
         iSection++)
    {
        pSection = *iSection;

        //
        // Grab the behavior from the section. In 7.1, this was on CustomSection.
        // In 7.2, we've promoted it to Section in general.
        //
        tSectionBehavior = pSection->behavior();

            //
            // manage section descriptor
            //
        if (tSectionBehavior.bPublishDescriptor)
        {
            //
            // name the new descriptor
            //
            zSection.assign( pSection->name() );
            zSection.append( /*NOXLATE*/L"\\descriptor.xml" );

            //
            // start with the descriptor
            //
            pFilestream = _pPackageDescriptor->zip( zSection, _zDWFPassword );

            //
            // bind the stream to the writer, this will directly
            // populate the archive as the page descriptor serializes itself
            //
            _pXMLSerializer->attach( *pFilestream );

            //
            // tell the page to serialize itself into the descriptor XML document
            //
            pSection->serializeXML( *_pXMLSerializer, eDescriptor );

            //
            // unbind the stream and finalize the file in the archive
            //
            _pXMLSerializer->detach();

            //
            // release the file stream
            //
            DWFCORE_FREE_OBJECT( pFilestream );
        }

        //
        // iterator for finders
        //
        DWFIterator<DWFResource*>* piResources = NULL;

            //
            // thumbnails
            //
        piResources = pSection->findResourcesByRole( DWFXML::kzRole_Thumbnail );
        if (piResources)
        {
            for (; piResources->valid(); piResources->next())
            {
                _serializeResource( pSection->name(), *(piResources->get()) );
            }

            //
            // clean up
            //
            DWFCORE_FREE_OBJECT( piResources );
            piResources = NULL;
        }

            //
            // fonts
            //
        piResources = pSection->findResourcesByRole( DWFXML::kzRole_Font );
        if (piResources)
        {
            for (; piResources->valid(); piResources->next())
            {
                _serializeResource( pSection->name(), *(piResources->get()) );
            }

            //
            // clean up
            //
            DWFCORE_FREE_OBJECT( piResources );
            piResources = NULL;
        }

            //
            // w2d
            //
        piResources = pSection->findResourcesByRole( DWFXML::kzRole_Graphics2d );
        if (piResources)
        {
            for (; piResources->valid(); piResources->next())
            {
                _serializeResource( pSection->name(), *(piResources->get()) );
            }

            //
            // clean up
            //
            DWFCORE_FREE_OBJECT( piResources );
            piResources = NULL;
        }

            //
            // w3d
            //
        piResources = pSection->findResourcesByRole( DWFXML::kzRole_Graphics3d );
        if (piResources)
        {
            for (; piResources->valid(); piResources->next())
            {
                _serializeResource( pSection->name(), *(piResources->get()) );
            }

            //
            // clean up
            //
            DWFCORE_FREE_OBJECT( piResources );
            piResources = NULL;
        }

            //
            // textures
            //
        piResources = pSection->findResourcesByRole( DWFXML::kzRole_Texture );
        if (piResources)
        {
            for (; piResources->valid(); piResources->next())
            {
                _serializeResource( pSection->name(), *(piResources->get()) );
            }

            //
            // clean up
            //
            DWFCORE_FREE_OBJECT( piResources );
            piResources = NULL;
        }

        //
        // remaining resources
        //
        DWFResource* pResource = NULL;
        DWFKVIterator<const wchar_t*, DWFResource*>* piAllResources = pSection->getResourcesByRole();

        if (piAllResources)
        {
            for (; piAllResources->valid(); piAllResources->next())
            {
                pResource = piAllResources->value();

                    //
                    // if someone can think of a faster, cleaner way of doing this, please fix it
                    //
                if ((pResource->role() != DWFXML::kzRole_Descriptor)    &&
                    (pResource->role() != DWFXML::kzRole_Thumbnail)     &&
                    (pResource->role() != DWFXML::kzRole_Font)          &&
                    (pResource->role() != DWFXML::kzRole_Graphics2d)    &&
                    (pResource->role() != DWFXML::kzRole_Graphics3d)    &&
                    (pResource->role() != DWFXML::kzRole_Texture))
                {
                    _serializeResource( pSection->name(), *pResource );
                }
            }

            DWFCORE_FREE_OBJECT( piAllResources );
        }
    }

        //
        // give extension chance to write a file after the sections
        //
    if (_pVersionExtension)
    {
       _pVersionExtension->postwriteSections( *this, *_pPackageDescriptor, *_pXMLSerializer, _zDWFPassword );
    }

    //
    // all done
    //
    _pPackageDescriptor->close();

        //
        // if we were aggregating, cleanup now
        //
    if (_bAggregating)
    {
        _postAggregationCleanup();
    }
}

void
DWFPackageWriter::_serializeContent( DWFContent* pContent )
throw( DWFException )
{
    if (_pPackageDescriptor == NULL)
    {
        _DWFCORE_THROW( DWFNullPointerException, /*NOXLATE*/L"No package descriptor exists" );
    }

    //
    // Determine the HREF to use
    //
    DWFString zHREF = pContent->href();

    //
    //  Create the output stream
    //
    DWFPointer<DWFOutputStream> apFilestream( _pPackageDescriptor->zip(zHREF, _zDWFPassword), false );

    //
    //  Get the input stream from the content
    //
    DWFPointer<DWFInputStream>  apInputStream( pContent->getInputStream(), false );

    //
    // sanity check
    //
    if (apFilestream.isNull() || apInputStream.isNull())
    {
        _DWFCORE_THROW( DWFUnexpectedException, /*NOXLATE*/L"Error acquiring streams" );
    }

    size_t  nBytesRead = 0;
    char    aBuffer[_DWFTK_PACKAGE_WRITER_RESOURCE_STREAM_BUFFER_BYTES];

    //
    // stream the bytes from the resource into the archive
    //
    while (apInputStream->available() > 0)
    {
        nBytesRead = apInputStream->read( aBuffer, _DWFTK_PACKAGE_WRITER_RESOURCE_STREAM_BUFFER_BYTES );
        apFilestream->write( aBuffer, nBytesRead );
    }

    //
    // release the file stream
    //
    apFilestream->flush();
}

void
DWFPackageWriter::_serializeResource( const DWFString&  zSection,
                                      DWFResource&      rResource )
throw( DWFException )
{
    if (_pPackageDescriptor == NULL)
    {
        _DWFCORE_THROW( DWFNullPointerException, /*NOXLATE*/L"No package descriptor exists" );
    }

    size_t  nBytesRead = 0;
    char    aBuffer[_DWFTK_PACKAGE_WRITER_RESOURCE_STREAM_BUFFER_BYTES];

    //
    // open the resource file in the archive
    //
    DWFString zHREF = rResource.publishedIdentity();

    //
    //
    //
    if (zHREF.bytes() == 0)
    {
        zHREF.assign( zSection );
        zHREF.append( /*NOXLATE*/L"\\" );
        zHREF.append( rResource.href() );
    }

    DWFPointer<DWFOutputStream> apFilestream( _pPackageDescriptor->zip(zHREF, _zDWFPassword), false );

    //
    // open the stream to acquire the resource bytes
    //
    DWFPointer<DWFInputStream>  apResourceStream( rResource.getInputStream(), false );

        //
        // sanity check
        //
    if (apFilestream.isNull() || apResourceStream.isNull())
    {
        _DWFCORE_THROW( DWFUnexpectedException, /*NOXLATE*/L"Error acquiring streams" );
    }

        //
        // stream the bytes from the resource into the archive
        //
    while (apResourceStream->available() > 0)
    {
        nBytesRead = apResourceStream->read( aBuffer, _DWFTK_PACKAGE_WRITER_RESOURCE_STREAM_BUFFER_BYTES );
        apFilestream->write( aBuffer, nBytesRead );
    }

    //
    // release the file stream
    //
    apFilestream->flush();
}

void
DWFPackageWriter::_ensureManifest()
throw( DWFException )
{
        //
        // create the manifest if necessary
        //
    if (_pPackageManifest == NULL)
    {
        //
        // when the manifest is created it is assigned a new UUID for it's object ID
        // all subsequent object IDs should be acquired via the next() method of the UUID generator
        //
        _pPackageManifest = DWFCORE_ALLOC_OBJECT( DWFManifest(_oUUID.uuid(true)) );
        if (_pPackageManifest == NULL)
        {
            _DWFCORE_THROW( DWFMemoryException, /*NOXLATE*/L"Failed to create manifest" );
        }
    }
}

_DWFTK_API
void
DWFPackageWriter::initializeForAggregation( teMergeContent eMergeType )
throw( DWFException )
{
    _bAggregating = true;
    _eMergeType = eMergeType;

    if (_oSections.size() > 0)
    {
        DWFContentManager* pContentManager = getContentManager();

        _DWFTK_STD_VECTOR(DWFSection*)::iterator iSection = _oSections.begin();
        for (; iSection != _oSections.end(); ++iSection)
        {
            _oSectionToContentManager.insert( std::make_pair( *iSection, pContentManager ) );
        }
    }
}

void
DWFPackageWriter::_aggregateContent()
throw( DWFException )
{
    //
    //  If we are aggregating pre 6.2 DWFs we don't need any content aggregation
    //  or merge. So lets count the number of associated contents first before we proceed.
    //
    DWFSortedVector<DWFString> oContentIDs( false );
    _tSectionToContentManager::iterator iMap = _oSectionToContentManager.begin();
    for (; iMap != _oSectionToContentManager.end();  ++iMap)
    {
        DWFSection* pSection = iMap->first;

        DWFSection::tStringMultiMap oResourceIDContentID;
        pSection->getAssociatedContentIDs( oResourceIDContentID );

        DWFSection::tStringMultiMap::iterator iResource = oResourceIDContentID.begin();
        for (; iResource != oResourceIDContentID.end(); ++iResource)
        {
            oContentIDs.insert( iResource->second );
        }
    }

    if (oContentIDs.empty())
    {
        return;
    }


    //
    //  Local content manager
    //
    DWFContentManager* pContentManager = getContentManager();

    //
    //  If we aren't merging just get all the contents from the other managers
    //  into this content manager.
    //
    if (_eMergeType == eNoMerge)
    {
        _tSectionToContentManager::iterator iMap = _oSectionToContentManager.begin();
        for (; iMap != _oSectionToContentManager.end(); ++iMap)
        {
            DWFSection* pSection = iMap->first;
            DWFContentManager* pCurrentManager = iMap->second;

            DWFSection::tStringMultiMap oResourceIDContentID;
            pSection->getAssociatedContentIDs( oResourceIDContentID );

            DWFSection::tStringMultiMap::iterator iResource = oResourceIDContentID.begin();
            for (; iResource != oResourceIDContentID.end(); ++iResource)
            {
                //
                //  First do a simple ID check - if a content with the given ID isn't in
                //  the output content manager, then pContent will be NULL. Hence
                //  grab the input content for aggregation
                //
                DWFString zContentID = iResource->second;
                DWFContent* pContent = pContentManager->getContent( zContentID );
                if (pContent == NULL)
                {
                    pContent = pCurrentManager->getContent( zContentID );
                    
                    //
                    //  Save the original owner to re-establish ownership after the write, where necessary
                    //
                    _oContentToContentManager.insert( std::make_pair( pContent, pContent->getContentManager() ) );

                    //
                    //  Now let the writer's content manager have it for the write
                    //
                    pContentManager->insertContent( pContent );
                }
            }
        }

        return;
    }

    //
    //  For merging, set whether the source or the target (primary) content element/property set
    //  has priority if a UUID or property name clashes.
    //
    bool bPrimaryHasPriority = (_eMergeType == eMergePriorityPrimary);

    //
    //  Get the primary content - this maybe empty - this is what everything will merge into
    //
    DWFContent* pPrimaryContent = pContentManager->getContent();

    //
    //  Don't want to revisit a content so track them in here
    //
    DWFSortedVector<DWFContent*> oVisitedContent( false );

    _tSectionVector::iterator iSection = _oSections.begin();
    for (; iSection != _oSections.end(); ++iSection)
    {
        DWFSection* pSection = (*iSection);

        //
        //  Skip this if the section has no corresponding content manager
        //
        _tSectionToContentManager::iterator iMap = _oSectionToContentManager.find( pSection );
        if (iMap == _oSectionToContentManager.end() || 
            iMap->second == NULL)
        {
            continue;
        }

        //
        //  This is the content manager to use to load section associated content
        //
        DWFContentManager* pCurrentManager = iMap->second;

        //
        //  The section no longer has the current manager as it's manager so 
        //  pSection->getContentDefinition will NOT work. 
        //
        DWFSection::tStringMultiMap oResourceIDContentID;
        pSection->getAssociatedContentIDs( oResourceIDContentID );

        DWFSection::tStringMultiMap::iterator iResource = oResourceIDContentID.begin();
        for (; iResource != oResourceIDContentID.end(); ++iResource)
        {
            DWFString zContentID = iResource->second;
            DWFContent* pContent = pCurrentManager->getContent( zContentID );

            if (pContent && 
                pContent != pPrimaryContent)
            {
                size_t iFirst, iLast;
                if (!oVisitedContent.find( pContent, iFirst, iLast ))
                {
                    pPrimaryContent->mergeContent( pContent, bPrimaryHasPriority );
                    oVisitedContent.insert( pContent );

                    //
                    //  If the content exists in the writer's content manager - delete the
                    //  content since its been merged in.
                    //
                    if (pCurrentManager == pContentManager)
                    {
                        pCurrentManager->removeContent( zContentID );
                    }
                }

                //
                //  Update the section's mapping of resource object IDs to content IDs. This will also
                //  update the corresponding resources.
                //
                pSection->updateResourceContentMapping( iResource->first, zContentID, pPrimaryContent->id() );
            }
        }
    }
}

void
DWFPackageWriter::_postAggregationCleanup()
throw()
{
    //
    //  Local content manager
    //
    DWFContentManager* pContentManager = getContentManager();

    _tContentToContentManager::iterator iContent = _oContentToContentManager.begin();
    for (; iContent != _oContentToContentManager.end(); ++iContent)
    {
        DWFContentManager* pCurrentManager = iContent->second;
        if (pCurrentManager &&
            pCurrentManager != pContentManager)
        {
            pCurrentManager->insertContent( iContent->first );
        }
    }

    _oSectionToContentManager.clear();
    _oContentToContentManager.clear();
    _bAggregating = false;
}

//DNT_End

#endif

