//
//  Copyright (c) 1996-2006 by Autodesk, Inc.
//
//  By using this code, you are agreeing to the terms and conditions of
//  the License Agreement included in the documentation for this code.
//
//  AUTODESK MAKES NO WARRANTIES, EXPRESS OR IMPLIED, AS TO THE CORRECTNESS
//  OF THIS CODE OR ANY DERIVATIVE WORKS WHICH INCORPORATE IT. AUTODESK
//  PROVIDES THE CODE ON AN "AS-IS" BASIS AND EXPLICITLY DISCLAIMS ANY
//  LIABILITY, INCLUDING CONSEQUENTIAL AND INCIDENTAL DAMAGES FOR ERRORS,
//  OMISSIONS, AND OTHER PROBLEMS IN THE CODE.
//
//  Use, duplication, or disclosure by the U.S. Government is subject to
//  restrictions set forth in FAR 52.227-19 (Commercial Computer Software
//  Restricted Rights) and DFAR 252.227-7013(c)(1)(ii) (Rights in Technical
//  Data and Computer Software), as applicable.
//

#ifndef _DWFTK_PACKAGE_WRITER_H
#define _DWFTK_PACKAGE_WRITER_H

///
///\file        dwf/package/writer/PackageWriter.h
///\brief       This file contains the DWFPackageWriter class declaration.
///


#ifndef DWFTK_READ_ONLY



#include "dwfcore/STL.h"
#include "dwfcore/UUID.h"
#include "dwfcore/Vector.h"
#include "dwfcore/OutputStream.h"
using namespace DWFCore;

#include "dwf/Toolkit.h"
#include "dwf/package/XML.h"
#include "dwf/package/Constants.h"
#include "dwf/package/writer/DWFXMLSerializer.h"
#include "dwf/package/writer/PackageVersionExtension.h"
#include "dwf/package/utility/DWFPackageFileDescriptor.h"

namespace DWFToolkit
{

//
// fwd decl
//
class DWFManifest;
class DWFResource;
class DWFInterface;
class DWFSection;
class DWFGlobalSection;
class DWFContentManager;
class DWFPropertyContainer;

///
///\ingroup     dwfpackage
///
///\class       DWFPackageWriter   dwf/package/writer/PackageWriter.h     "dwf/package/writer/PackageWriter.h"
///\brief       This class is used to publish DWF package files.
///\since       7.0.1
///
///\todo        Add direct publishing to DWFOutputStream in addition to DWFFile respresentation.
///
class DWFPackageWriter      _DWFTK_EXPORT_API_MEMORY_CONTROL_BASE
{

public:

        ///
        ///\brief   These enumerations are used during the serialization process
        ///         to alter the content produced by the serializable objects.
        ///
    typedef enum teDocumentType
    {
        eManifest               = 0x0010,
        eContentPresentation    = 0x0020,
        eDescriptor             = 0x0040,
        eObjectDefinition       = 0x0080,
        eGlobalContent          = 0x0100,
        eSectionContent         = 0x0200
    } teDocumentType;

        ///
        ///\brief   These enumerations are used during the serialization process
        ///         to alter the content produced by the serializable objects.
        ///
    typedef enum teSectionType
    {
        eNone               = 0x00,
        eEPlot              = 0x01,
        eEModel             = 0x02,
        eGlobal             = 0x04,
        eData               = 0x08

    } teSectionType;

    typedef enum teMergeContent {
        eNoMerge,
        eMergePriorityPrimary,
        eMergePrioritySecondary
    } teMergeContent;

public:

    ///
    ///         Constructor
    ///
    ///\param   rDWFPackageFile         Identifies the DWF file to publish.
    ///\param   zDWFPackagePassword     An optional password for protecting the DWF package file.
    ///\param   pVersionExtension       An optional plug-in for publishing package versions different from the curent one.
    ///\throw   None
    ///
    _DWFTK_API
    DWFPackageWriter( const DWFFile&                rDWFPackageFile,
                      const DWFString&              zDWFPackagePassword = /*NOXLATE*/L"",
                      DWFPackageVersionExtension*   pVersionExtension = NULL )
        throw();

    ///
    ///         Destructor
    ///
    ///\throw   None
    ///
    _DWFTK_API
    virtual ~DWFPackageWriter()
        throw();

    ///
    ///         Adds a section to the DWF package.
    ///
    ///\param   pSection                The section to publish into the package.
    ///                                 Ownership of this section pointer will be claimed automatically.
    ///                                 This means that, left as is, the pointer will be deleted
    ///                                 once the package is published.  
    ///                                 In order to prevent this, the caller can explicitly reclaim
    ///                                 ownership of the section by calling DWFSection::own().
    ///
    ///\param   pInterface              This object should be provided if \a pSection is a custom type that
    ///                                 should be declared in the manifest (it will be ignored for known types.)
    ///                                 Note that the DWF Specification dictates that at least one interface
    ///                                 be declared in the manifest.  If none exist, an exception will be thrown at write time.
    ///                                 This pointer will be deleted with the \b DWFCORE_FREE_OBJECT macro.
    ///\throw   DWFException
    ///
    _DWFTK_API
    void addSection( DWFSection*    pSection,
                     DWFInterface*  pInterface = NULL )
        throw( DWFException );

    ///
    ///         Adds a type global section to the DWF package.
    ///
    ///\param   pSection                The global section to publish into the package.
    ///                                 Ownership of this section pointer will be claimed automatically.
    ///                                 This means that, left as is, the pointer will be deleted
    ///                                 once the package is published.  
    ///                                 In order to prevent this, the caller can explicitly reclaim
    ///                                 ownership of the section by calling \a DWFSection::own().
    ///\throw   DWFException
    ///
    _DWFTK_API
    void addGlobalSection( DWFGlobalSection* pSection )
        throw( DWFException );

    ///
    ///         Adds an interface to the DWF package manifest.
    /// 
    ///         This method is provided in addition to the \a addSection() method
    ///         as it may be necessary to, in the case of dumb package copying
    ///         for example, to include this information in the target package
    ///         without having any ability to corelate the source sections
    ///         and interfaces.  
    ///         This pointer will be deleted with the \b DWFCORE_FREE_OBJECT macro.
    ///
    ///         This method will check for duplicate interfaces and
    ///         if \a pInternface is a duplicate, it will
    ///         be deleted and the previously stored one will be returned.
    ///
    ///\throw   DWFException
    ///
    _DWFTK_API
    DWFInterface* addInterface( DWFInterface* pInterface )
        throw( DWFException );

    ///
    ///         This attaches the content manager to the manifest owned
    ///         by the package writer. The manifest can take ownership of
    ///         the manager if requested to.
    ///
    ///\param   pContentManager         The content manager to attach.
    ///\param   bTakesOwnership         Set this to true if the ownership is supposed change.
    ///\throw   DWFException
    ///
    ///\since   7.2.0
    ///
    _DWFTK_API
    void attachContentManager( DWFContentManager* pContentManager, bool bTakesOwnership = true )
        throw( DWFException );

    ///
    ///         This returns the content manager associated with this writer.
    ///         If a content manager was not attached to the writer, a new one
    ///         will be created by the writer's manifest and returned here. Note
    ///         the caller should not delete the content manager unless it was attached
    ///         using \a attachContentManager with \a bTakeOwnership set to false.
    ///
    ///\return  Pointer to the content manager.
    ///\throw   DWFException
    ///
    ///\since   7.2.0
    ///
    _DWFTK_API
    DWFContentManager* getContentManager()
        throw( DWFException );

    ///
    ///         Returns the property container interface
    ///         for the package manifest.  
    ///
    ///         This can be used to add package-wide properties to the DWF.
    ///
    ///\return  An interface with which to add package manifest properties.
    ///\throw   DWFException
    ///
    _DWFTK_API
    DWFPropertyContainer& getManifestProperties()
        throw( DWFException );

    ///
    ///         Initialize the writer for aggregation. This lets the writer collect
    ///         information required for aggregation as the sections are added.
    ///         This should be called right after the writer has been created.
    ///
    ///\param   eMergeType      This defines whether or not to merge the content if there
    ///                         multiple contents.
    ///\throw   None
    ///
    ///\since   7.2.0
    ///
    _DWFTK_API
    void initializeForAggregation( teMergeContent eMergeType = eMergePriorityPrimary )
        throw( DWFException );

    ///
    ///         Writes out the DWF package.
    ///
    ///         Any sections and other objects owned by the writer and/or internal manifest
    ///         will be deleted at the completion of this method.  Be sure to claim ownership
    ///         of any such objects that are necessary to survive after this call.
    ///
    ///         This method exposes the standard DWF package properties directly.
    ///         It is highly recommended that at DWF publishers set these correctly with as
    ///         much detail as possible.
    ///
    ///         Setting standard DWF package properties:
    ///         \code
    ///         Autodesk DWF Writer is a printer driver DLL that is used to generate DWF packages
    ///         through the Windows printing system, it is a component that another application
    ///         would use (indirectly) to publish the DWF.  The following is an example of how
    ///         it might populate these properties when printing from Microsoft Word:
    ///
    ///         zSourceProductVendor    = L"Microsoft Corporation"          <--- from Word 2003
    ///         zSourceProductName      = L"Microsoft Office 2003"          <--- from Word 2003
    ///         zSourceProductVersion   = L"11.0.5604"                      <--- from Word 2003
    ///         zDWFProductVendor       = L"Autodesk, Inc."                 <--- from DWF Writer 2
    ///         zDWFProductVersion      = L"2.0.4.0"                        <--- from DWF Writer 2
    ///         \endcode
    ///
    ///\param   zSourceProductVendor    A standard property detailing the vendor of the software
    ///                                 that produced the original content and used the DWF product
    ///                                 in order to produce the DWF package.
    ///\param   zSourceProductName      A standard property detailing the name of the software
    ///                                 that produced the original content and used the DWF product
    ///                                 in order to produce the DWF package.
    ///\param   zSourceProductVersion   A standard property detailing the version of the software
    ///                                 that produced the original content and used the DWF product
    ///                                 in order to produce the DWF package.
    ///\param   zDWFProductVendor       A standard property that details the vendor of the software component
    ///                                 that was used by the source product to publish the DWF package.
    ///\param   zDWFProductVersion      A standard property that details the version of the software component
    ///                                 that was used by the source product to publish the DWF package.
    ///\param   eCompressionMode        An enumeration that adjusts the zip library compression options.
    ///
    ///\throw   DWFException
    ///
    _DWFTK_API
    void write( const DWFString&                 zSourceProductVendor = /*NOXLATE*/L"",
                const DWFString&                 zSourceProductName = /*NOXLATE*/L"",
                const DWFString&                 zSourceProductVersion = /*NOXLATE*/L"",
                const DWFString&                 zDWFProductVendor = /*NOXLATE*/L"",
                const DWFString&                 zDWFProductVersion = /*NOXLATE*/L"",
                DWFZipFileDescriptor::teFileMode eCompressionMode = DWFZipFileDescriptor::eZipSmallest )
        throw( DWFException );

private:

    //
    //  This is used only in cases where the content is available but not loaded - may happen
    //  during aggregation.
    //
    void _serializeContent( DWFContent* pContent )
        throw( DWFException );

    //
    //
    //
    void _serializeResource( const DWFString&   zSection,
                             DWFResource&       rResource )
        throw( DWFException );

    //
    // 
    //
    void _ensureManifest()
        throw( DWFException );

    //
    //  This will aggregate content from other packages, and merge if the merge flag is set
    //
    void _aggregateContent()
        throw( DWFException );

    //
    //  If aggregation was performed this will cleanup, and more important - restore
    //  contents to their original owner content managers.
    //
    void _postAggregationCleanup()
        throw();

private:
    
    typedef _DWFTK_STD_VECTOR(DWFSection*)                  _tSectionVector;
    typedef _DWFTK_STD_VECTOR(DWFGlobalSection*)            _tGlobalSectionVector;
    typedef _DWFTK_STD_MAP(DWFSection*, DWFContentManager*) _tSectionToContentManager;
    typedef DWFSortedVector<DWFContentManager*>             _tSortedContentManagerVector;
    typedef _DWFTK_STD_MAP(DWFContent*, DWFContentManager*) _tContentToContentManager;

private:

    DWFFile                         _rDWFPackage;
    DWFString                       _zDWFPassword;

    DWFUUID                         _oUUID;
    DWFXMLSerializer*               _pXMLSerializer;
    DWFPackageVersionExtension*     _pVersionExtension;

    DWFPackageFileDescriptor*       _pPackageDescriptor;
    DWFManifest*                    _pPackageManifest;

    double                          _nNextPlot;

    bool                            _bEPlotGlobalAdded;
    bool                            _bEModelGlobalAdded;

    size_t                          _nEPlotSectionCount;
    size_t                          _nEModelSectionCount;

    _tSectionVector                 _oSections;
    _tGlobalSectionVector           _oGlobalSections;

    bool                            _bAggregating;
    teMergeContent                  _eMergeType;

    //
    //  The section and the "original" content manager that it was associated with before being added
    //
    _tSectionToContentManager       _oSectionToContentManager;

    //
    //  The contents that get aggregated and the "original" content managers they came from
    //
    _tContentToContentManager       _oContentToContentManager;

private:

    DWFPackageWriter();
    DWFPackageWriter( const DWFPackageWriter& );
    DWFPackageWriter& operator=( const DWFPackageWriter& );
};

}

#endif
#endif

